/*  FINAL PROJECT: SNORE DETECTOR *\
*             Da, Jimmy            *
*           Meller, Aaron          *
*                                  *
* Master Port List:                *
*   D2-4: buttons                  *
*   D5: LED (for SD writes)        *
*   C7-0: LCD screen               *
*   B7-5: SPI                      *
*   B4: SPI CS                     *
*   A0: ADC mic input              *
*                                  *
***********************************/
#include <inttypes.h> //ADC LIBRARIES
#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdio.h>
#include <math.h>
#include "uart.h"

#include <avr/pgmspace.h> //LCD LIBRARIES
#include <stdlib.h>
#include <string.h>
#include "lcd_lib.h"

#include <util/delay.h>    //SD LIBRARIES
#include <stddef.h>
#include "main.h"
#include "fatfs/ff.h"

//set up ASSERT
#ifndef __ASSERT_USE_STDERR
#define __ASSERT_USE_STDERR
#include <assert.h>
#endif

FILE uart_str = FDEV_SETUP_STREAM(uart_putchar, uart_getchar, _FDEV_SETUP_RW);

#define DEBUG

#define t1 200 //timeout value for LCD task, ms
#define tb 35 //timeout value for button task, ms
#define ts 1000 //timeout value for timeStamp
#define diskInterval 10  //Interval for disk_timerproc (NEEDS to be 10ms)
#define countAvg 127 // The number of samples to average for output
#define debugInterval 1000 //1s for debug

//sd setup
volatile int debugTime;

FIL logFile;
FIL* logPtr;
FATFS fs;
FATFS* fsPtr;

void disk_timerproc(); //declaration for disk_timerproc(), in ff.c

// clock setup
volatile unsigned int msCount, t1Count;// ms counter, in timer0 isr
volatile unsigned long timeStamp, timeStamp_halt, timeStamp_cur; //second timer since start, in timer0 isr
int sec_s; //minute timing for clock calculation
char num_ball, hrs, mins, secs; //values for clock

//LCD SETUP ************************************************
const int8_t LCD_initialize[] PROGMEM = "LCD Initialized\0";
const int8_t LCD_line[] PROGMEM = "Time\0";

int8_t lcd_buffer[17], lcd_buffer1[17];    // LCD display buffer

volatile char diskTime;      //disk_timerproc timer
volatile unsigned int time1, time2, timeButton;// running time counter for task1 

//Button stuff
//State machine state names
#define NoPush        1
#define MaybePush    2
#define Pushed        3
#define MaybeNoPush 4

//Record mode
#define TRIGGER        0
#define CONTINUOUS    1

//Button stuff
// shared led status
uint8_t led ;
uint8_t sw, sw_num, maybe ;
uint8_t PushState, PushFlag ;

//float-fix conversion macros
#define float2fix(a) ((int)((a)*256.0))
#define fix2float(a) ((double)(a)/256.0)

// Fast fixed point multiply assembler macro
#define multfix(a,b)                \
({                                \
int prod, val1=a, val2=b ;        \
__asm__ __volatile__ (            \
"muls %B1, %B2    \n\t"              \
"mov %B0, r0 \n\t"                   \
"mul %A1, %A2\n\t"                   \
"mov %A0, r1   \n\t"              \
"mulsu %B1, %A2    \n\t"          \
"add %A0, r0  \n\t"               \
"adc %B0, r1 \n\t"                \
"mulsu %B2, %A1    \n\t"          \
"add %A0, r0 \n\t"           \
"adc %B0, r1  \n\t"          \
"clr r1  \n\t"                  \
: "=&d" (prod)               \
: "a" (val1), "a" (val2)      \
);                            \
prod;                        \
})


//IIR state variables
int b1,b2,b3,a2,a3, a4, a5 ;
int xn, xn_1, xn_2, xn_3, xn_4 ;
int yn_1, yn_2, yn_3, yn_4  ;

//CIC state variables
long integrator1, integrator2, integrator3, integrator4 ;
long last_integrator4, last2_integrator4 ;
long last_integrator2, last2_integrator2 ;
long comb1, comb2, comb3, comb4 ;
long last_comb1, last_comb2, last_comb3, last_comb4 ;
long last2_comb1, last2_comb2, last2_comb3, last2_comb4 ;
char downsample_clock ;
char Ain;
int yy ;
int xx ;

// IIr state variables
int b1,b2,b3,a2,a3;
int xn, xn_1, xn_2 ;
int yn_1, yn_2 ;
unsigned int zz, zz0, zz1, zz2, zzNow ;
unsigned int accumulator;
unsigned char zzTh, zzHigh, zzAHA, zzAvg ;

unsigned char count, countLog; // keep a running average for accumulator
char flag, record; //flag signaling just activate, record controls record mode
char snoreFlag, snoring, contFlag; //signaling snore is detected
char intensityFlag; //Intensity of snore is obtained
char activate; //signaling on/off is pushed and start recording
 
//******************************************************
//timer 0 compare ISR **********************************
ISR (TIMER0_COMPA_vect)
{   
    // read a/d converter
    xx = ((int)(abs(ADCH-100))<<8) ;
    // start new a/d conversion
    ADCSRA |= (1<<ADSC) ;
   
    // lowpass and downsample
    CIC_n4_r4_m2(xx) ;

    // update PWM with Cheby1 filter output
    OCR2A = (char)(zz>>8) ;
    if (OCR2A == 255) OCR2A = 0;
    accumulator += OCR2A;
    count--;

    if (activate && count==0 && record==TRIGGER) {
        count = countAvg;
        zzAvg = accumulator>>7;

        accumulator = 0;
        if(zzAvg>zzTh && !snoring) {
            zzHigh = zzAvg;
            snoreFlag = 1;
            snoring = 1;
            timeStamp_cur = timeStamp;
            timeStamp_halt = timeStamp + 2; 
        }
      
        if(snoring && zzAvg>zzHigh) {
            zzHigh = zzAvg;
        }
      
        if(snoring && (timeStamp==timeStamp_halt)) {
            snoring = 0;
            intensityFlag = 1;
            zzAHA = zzHigh;
            zzHigh= 0;
        }
    }
}

//timer 1 compare ISR **********************************
ISR (TIMER1_COMPA_vect) //ms tick
{
   
    ticks++;
    contFlag = 1;  
    //decrement time
    if (diskTime>0) --diskTime;
    if (diskTime == 0) {
            diskTime = diskInterval;
            disk_timerproc();
    }
   
    if (time1>0) time1--; // updating LCD

    if (timeButton>0) timeButton--;  //scan buttons

    if (time2>0) time2--; // incrementing seconds
    else {
        time2=ts;
        timeStamp++;
    }

    IncrementSysTime();
}


// TASKS ***********************************************
//Update the current system in days/hours/minutes at a 1ms
void IncrementSysTime() {
    if (++msec == 60000) {
        msec = 0;
        minutes++;
    }
    if (minutes == 60) {
        minutes = 0;
        hours++;
    }
    if (hours == 24) {
        hours = 0;
        days++;
    }
}


int main(void) {
    initialize();
    init_fs(); //init the  FAT file system
/*    open_log();    //Open a new log file a log file
    #ifdef DEBUG
    fprintf(stdout, "openlog has run\r\n");
    #endif
*/   
   //main task scheduler loop
      while(1)
      {

      // reset time and call task
        //runs every 200 ms
        if (time1==0 && !activate){
             time1=t1;
             lcdWrite();
         }

         //runs every 50ms
         if (timeButton==0) {
            timeButton=tb;
            debouncer();
            if (flag) {
                open_log();
                flag = 0;
            }
            //fprintf(stdout, "%d \n\r", zzAvg);
         }

           switch (record){
            case TRIGGER:
                if(snoreFlag) {
                    snoreFlag = 0;
                    fprintf(stdout, "snore detected at %d ", timeStamp_cur);
                }
                if(intensityFlag) {
                    intensityFlag = 0;
                    fprintf(stdout, "with intensity of %d/255 \n\r", (zzAHA));
                    zzNow = zzAHA;
                    writeLog(); //log the snore
                    f_sync(logPtr);
                   
                   
                }
                break;
            case CONTINUOUS:
                if(activate && contFlag) {
                    contFlag = 0;
                    writeLog(); //log the snore
                }
                break;
        }
      }
      return 0;
}

//************Button*************************************
// Button Debouncing
void debouncer(void) {
    sw = (~PIND) & 0b00111100;
/* 
    led = sw;
    PORTB = ~(led & 0b00111100) ;
*/
    switch (PushState) {
        case NoPush:
            if (sw !=0) {
                PushState = MaybePush;
                maybe = sw;
                PushFlag = 0;
            }
            else {
                PushState = NoPush;
                sw = 0;
            }
            break;
     
        case MaybePush:
            if (sw == maybe) {
                PushState = Pushed;
                PushFlag=1;
                sw_num = 0 ;
                while(sw>1) {
                    sw = sw>>1;
                    sw_num++ ;
                }
                sw_num--;
                buttonFun(sw_num);
                #ifdef DEBUG
                fprintf(stdout,"Button pushed=%d  ", sw_num) ;
                fprintf(stdout,"Activated=%s  ", activate?"on":"off") ;
                fprintf(stdout,"Threshold=%d\n\r", zzTh) ;
                #endif
            }
            else PushState = NoPush;
            break;

        case Pushed:
            if (sw == maybe) {
                PushState=Pushed;
            }
            else PushState=MaybeNoPush;
            break;
        case MaybeNoPush:
            if (sw == maybe) {
                PushState=Pushed;
            }
            else {
               PushState=NoPush;
               PushFlag=0;
            }
            break;
         
        default: PushState=NoPush; 
    }
}
// Button function
void buttonFun(uint8_t sw_num) {
    switch (sw_num) {
        case 1: // on/off for recording
            activate = ~activate;
            if (!activate) {
                snoring = 0;
                f_close(logPtr);
            }
            if (activate) {
                PORTC = 0x00;
                flag = 1;
            }
            break;
          
        case 2: // increase threshold
            if(zzTh<255) zzTh += 17;
            else zzTh = 255;
            break;
          
        case 3: // decrease threshold
            zzTh -= 17;
            if(zzTh<20) zzTh = 17;
            break;
          
        case 4: // change recording mode
            if(!activate)    record = 1-record;
            break;

        default: break;
    }
}

//*************DSP***************************************
// Second order Butterworth lowpass
// NOTE chebshev1 type filters also work with this code
// xx is the current input signal sample
// returns the current filtered output sample
// Simplification: b(1)*x(n)+b(2)*x(n-1)+b(3)*x(n-2) =
// b(1)*x(n)+2*b(1)*x(n-1)+b(1)*x(n-2) =
// b(1)* (x(n)+(x(n-1)<<1)+x(n-2))
int Butter2low(int xx) {
    int yy  ;
    // sum the 5 terms: yy += xx*coeff
    // and update the state variables
    // as soon as possible
 
    yy = multfix(xx+(xn_1<<1)+xn_2, b1);
    xn_2 = xn_1;
    xn_1 = xx;
    yy = yy + multfix(yn_2,a3);
    yn_2 = yn_1;
    yy = yy + multfix(yn_1,a2);
    yn_1 = yy;
    return yy;
}
// 4th order CIC
// takes 356 cycles when downsampling
// takes 172 cycles when just integrating
void CIC_n4_r4_m2(int xx) {
     integrator1 = integrator1 + xx ;
     integrator2 = integrator2 + integrator1 ;
     integrator3 = integrator3 + integrator2 ;
     integrator4 = integrator4 + integrator3 ;
   
     downsample_clock++ ;
     if ((downsample_clock & 0x0B) == 0) //modulo 12
     {
        comb1 = integrator4 - last2_integrator4 ;
        comb2 = comb1 - last2_comb1 ;
        comb3 = comb2 - last2_comb2 ;
        yy = (int)((comb3 - last2_comb3)>>16) ;
     
        // update state
        last2_integrator4 = last_integrator4 ;
        last_integrator4 = integrator4 ;
        last2_comb1 = last_comb1 ;
        last_comb1 = comb1;
        last2_comb2 = last_comb2 ;
        last_comb2 = comb2;
        last2_comb3 = last_comb3 ;
        last_comb3 = comb3;
     }
   
     // on the next phase of the clock do
     // the chebyshev compensation
     if ((downsample_clock & 0x0B) == 1) //modulo 12
         zz = Butter2low(yy) ;
}

//*************LCD***************************************
// task writes to LCD every 200 mSec
void lcdWrite (void)
{

    // calculate clock time from timeStamp
    hrs = floor(timeStamp/3600);
    mins = floor( (timeStamp - (3600*hrs)) / 60);
    secs = timeStamp - (3600*hrs) - (60*mins);//not necessary to output

    //print current clock time
    sprintf(lcd_buffer,"%02d:%02d:%02d  %s", hrs, mins, secs, record?"Cont":"Trig");                  
    LCDGotoXY(0, 0);
    LCDstring(lcd_buffer, strlen(lcd_buffer));

    //show threshold
    num_ball = (yy>>8)/7+1;
    LCDGotoXY(0, 1);
    strcpy(lcd_buffer1,"");
    for(int i=0; i<17; i++) {
      if(i<num_ball)    strcat(lcd_buffer1,">");
    else            strcat(lcd_buffer1," ");
    }
    LCDstring(lcd_buffer1, strlen(lcd_buffer1));

    num_ball=0;
}

//***********FatFs and SD card writing*******************
// Implemented to enable FatFs writing
DWORD get_fattime (void) {
    /*
    bit31:25
    Year from 1980 (0..127)
   
    bit24:21
    Month (1..12)
   
    bit20:16
    Day in month(1..31)
   
    bit15:11
    Hour (0..23)
   
    bit10:5
    Minute (0..59)
   
    bit4:0
    Second / 2 (0..29)
    */
   
    // In order to have f_write function to work well,
    // get_fattime needed to be implemented, even though
    // this is not entirely accurate
    return      ((DWORD)(2011 - 1980) << 25)
            | ((DWORD)5 << 21)
            | ((DWORD)(days+1) << 16)
            | ((DWORD)hours << 11)
            | ((DWORD)minutes << 5)
            | ((DWORD)(msec >> 10) >> 1);
    // div 1024 is close enough to div 1000 since the time is
    // pretty made up in this case anyway

   
}
// Creating a new file
void open_log(void) {
    FRESULT result;
    char filePath[16];
    unsigned int counter = 0;
    //Try to create a new log file. It will fail if the log exists, keep incrementing log file counter until we succeed
    switch (record){
        case TRIGGER:
            do {
                snprintf(filePath, 16, "sTrig%03d.csv", counter);
                result = f_open(logPtr, filePath, FA_WRITE | FA_CREATE_NEW);
            } while (result == FR_EXIST && ++counter < 1000);
            if(f_printf(logPtr, "Time, Seconds, Strength\n")>0)
                fprintf(stdout, "Log write succeeded.\r\n");
            else    fprintf(stdout, "Log write error.\r\n");
            break;
        case CONTINUOUS:
            do {
                snprintf(filePath, 16, "sCont%03d.csv", counter);
                result = f_open(logPtr, filePath, FA_WRITE | FA_CREATE_NEW);
            } while (result == FR_EXIST && ++counter < 1000);
            if(f_printf(logPtr, "mSeconds, Strength\n")>0)
                fprintf(stdout, "Log write succeeded.\r\n");
            else    fprintf(stdout, "Log write error.\r\n");
            break;
    }
    if (result) {
        #ifdef DEBUG
        fprintf(stdout, "Snore log creation FAILED: %d\r\n", result);
        #endif
    }
    else {
        #ifdef DEBUG
        fprintf(stdout, "Snore Log Created: %s\n\r", filePath);
        #endif
    }

}
// Writing to the SD card
void writeLog(void) {
  
    //turn on the LED
    //indicates a write, and shows that filesystem is in
    // a critical section
//    PORTD &= ~(1 << PORTD5);

    switch (record){
    case TRIGGER:
        fprintf(stdout, "%d  %d\n\r", zzNow, zzAHA);
        if(f_printf(logPtr, "%02d:%02d:%02d, %d, %d \n", hrs, mins, secs, timeStamp_cur, (zzNow))>0)
            break;
        else    fprintf(stdout, "Log write error.\r\n");
        break;
   
    case CONTINUOUS:
        if(f_printf(logPtr, "%lu, %d \n", ticks, ADCH)>0)
            break;
        else    fprintf(stdout, "Log write error.\r\n");
        break;
    //sync the filesystem (flush buffers) so it will power down cleanly
    }
}

//***********Initializations*****************************
void initialize(void) {

   // ADC Stuff ********************************************
   ADMUX = (1<<ADLAR) ;
   ADCSRA = ((1<<ADEN) | (1<<ADSC)) + 7 ;
   uart_init();   // init the UART -- uart_init() is in uart.c
   stdout = stdin = stderr = &uart_str;
   fprintf(stdout,"Starting ADC power...\n\r");
   // ADC Stuff ********************************************
 
   // LCD Stuff ********************************************
    // start the LCD
    init_lcd();
    LCDclr();

    //init the ADC timer
    // time1=t1;
    //set up timer 0 for 8kHz tick
    TIMSK0 = 2;        //turn on timer 0 cmp match ISR
    OCR0A = 249;      //set the compare reg to 250 time ticks
    TCCR0A = 0b00000010; // turn on clear-on-match
    TCCR0B = 0b00000010;    // clock prescalar to 8, clk=2MHz
 
    //set up timer 1 for 1 mSec ticks
    OCR1A = 249;  //clear after 250 counts
    TIMSK1 = (1<<OCIE1A); //turn on timer 1 compare match ISR
    //turn on timer1 clear on match
    TCCR1A = (1 << COM1A1);
    //timer 0 prescalar to 64
    TCCR1B = 3|(1 << WGM12);
/* 
    // timer 2 PWM on, full speed
    TCCR2B = 1;
    TCCR2A =  (1<<COM2A1) | (1<<WGM20) | (1<<WGM21) ; //(1<<COM0A0) |
    // make D.7 an output for OC2A
    DDRD = (1<<PORTD7) ;
     TIMSK2 = 0 ;
*/ 
    downsample_clock = 0 ; // zero is worst case time

    // 2-pole cheby1 lowpass, [b,a] = cheby1(2, 8, 0.5)
    // b = 0.1176    0.2352    0.1176
    // a = 1.0000   -0.4931    0.6746
    b1 = float2fix(0.1176) ;
    b2 = float2fix(0.2352) ;
    b3 = float2fix(0.1176) ;             
    a2 = float2fix(0.4931) ;   //note that sign is negated from design input
    a3 = float2fix(-0.6746) ;  //note that sign is negated from design input

   
    // put some stuff on LCD
//    CopyStringtoLCD(LCD_line, 0, 0);//start at char=8 line=1 
    //CopyStringtoLCD(LCD_number, 0, 0);//start at char=0 line=0
   // LCD Stuff ********************************************

   // Clock Stuff
   msCount = timeStamp = sec_s = 0; //ms counter, timeStamp counter from start
   hrs =mins=secs =0; //initial hours, mins, secs for clock on LCD
   t1Count = 0;
   time1 = t1;
   time2 = ts;
   timeButton = tb;
  
   ticks = 0L;
 
    //init the  message
    //for no button push
    PushFlag = 0;
    //init the state machine
    PushState = NoPush;
  
    zzTh = 136; // start in the middle
    zzHigh = zz = 0;
    count = countAvg;

    flag = 0;
    snoreFlag = intensityFlag = snoring = activate = 0;

    timeStamp_halt = 0;
    // Button stuff
//    DDRD |= 0x00; //leave top 2 bits for PuTTy
    PORTD|= (1<<PORTD2)|(1<<PORTD3)|(1<<PORTD4)|(1<<PORTD5);
 //   DDRB = 0xff; //delete before implement SPI tests button push
 
   sei();
}
// File system setup
void init_fs() {
    FRESULT result;
  
    //initialize the pointers
    fsPtr = &fs;
    logPtr = &logFile;
    //set up the file system
    result = f_mount(0, fsPtr);
    if (result) {
        #ifdef DEBUG
        fprintf(stdout, "Filesystem mount FAILED!\r\n");
        #endif
        return;
    }
}
// LCD setup
void init_lcd(void) {
        LCDinit();    //initialize the display
        LCDcursorOFF();
        LCDclr();                //clear the display
        LCDGotoXY(0,0);
        CopyStringtoLCD(LCD_initialize, 0, 0);
} 
